#include "pointsetdisplay.h"
#include <QMouseEvent>
#include <iostream>
#include <math.h>
#include "util.h"

PointSetDisplay::PointSetDisplay(QWidget *parent) :
        DGLWidget(parent, QGLFormat(QGL::DepthBuffer | QGL::DoubleBuffer))
{
    _set = 0;
    _translate = _rotate = _zoom = false;
    _faces = 0;
    _nfaces = 0;
    _hilightlast = false;
}

PointSetDisplay::~PointSetDisplay()
{
    setPointSet(0); // deletes all points and faces
}

void PointSetDisplay::initializeGL()
{
    DGLWidget::initializeGL();
    qglClearColor(Qt::white);
    glClearDepth(1.0f);
    glPointSize(3.0f);
    glEnable(GL_DEPTH_TEST);
}

void PointSetDisplay::setPointSet(PointSet *set)
{
    if (set) set->ref();

    if (_set) {
        _set->unref(); // don't own the old point set anymore
        clearFaces();
    }

    _set = set;
    if (_set) {

        if (_set->size() > 0) {
            Point *p;

            p = _set->points();
            _minx = _maxx = p[0].x();
            _miny = _maxy = p[0].y();
            _minz = _maxz = p[0].z();

            for (int i = 1; i < _set->size(); ++i) {
                if (p[i].x() < _minx) _minx = p[i].x();
                if (p[i].x() > _maxx) _maxx = p[i].x();
                if (p[i].y() < _miny) _miny = p[i].y();
                if (p[i].y() > _maxy) _maxy = p[i].y();
                if (p[i].z() < _minz) _minz = p[i].z();
                if (p[i].z() > _maxz) _maxz = p[i].z();
            }
        }
        else {
            _minx = _maxx = 0.0;
            _miny = _maxy = 0.0;
            _minz = _maxz = 0.0;
        }
    }

    _viewinit = false;
}

void PointSetDisplay::clearFaces()
{
    free(_faces);
    _faces = 0;
    _nfaces = 0;
}

void PointSetDisplay::addFace(const Point *p, const Point *q, const Point *r)
{
    _faces = (Face*) realloc(_faces, sizeof(Face) * (++_nfaces ));
    _faces[_nfaces-1].set(p, q, r);
}

void PointSetDisplay::resizeGL(int w, int h)
{
    DGLWidget::resizeGL(w, h);

    glViewport(0, 0, w, h);

    _top = -1.0*h/w;
    _bottom = 1.0*h/w;
    _left = -1.0;
    _right = 1.0;
    _near = 1.0;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(_left, _right, _top, _bottom, _near, 2000.0);
    glMatrixMode(GL_MODELVIEW);
}

void PointSetDisplay::paintGL()
{
    if (!_viewinit) initView();

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glPushMatrix();

    // position the camera based on user input
    glTranslatef(_posx, _posy, _posz);
    glRotatef(_rotx, 0.0f, 1.0f, 0.0f);
    glRotatef(_roty, cos(rad(_rotx)), 0.0f, sin(rad(_rotx)));

    //glRotatef(90.0f, 0.0f, 0.0f, -1.0f);

    // center it all on the center of the point set instead of the origin
    glTranslatef((_minx - _maxx) / 2.0 - _minx,
                 (_miny - _maxy) / 2.0 - _miny,
                 (_minz - _maxz) / 2.0 - _minz);

    // draw the points
    if (_set) {
        Point *p;
        register int i, size;
        float cred, cgreen, cblue;

        p = _set->points();
        size = _set->size();

        cred = cgreen = cblue = 0.0f;
        glColor3f(cred, cgreen, cblue);

        glBegin(GL_POINTS);
        for (i = 0; i < size; ++i) {
            if (p[i].red != cred || p[i].green != cgreen || p[i].blue != cblue) {
                cred = p[i].red;
                cgreen = p[i].green;
                cblue = p[i].blue;
                glColor3f(cred, cgreen, cblue);
            }
            glVertex3dv(p[i].glcoords);
        }
        glEnd();
    }

    if (_faces) {
        Face *f;
        register int i, j, size;
        float cred, cgreen, cblue;

        f = _faces;
        size = _nfaces;

        cred = cgreen = cblue = 0.0f;
        glColor3f(cred, cgreen, cblue);

        glBegin(GL_TRIANGLES);
        for (i = 0; i < size; ++i) {
            for (j = 0; j < 3; ++j) {
                if (i == size - 1 && _hilightlast) {
                    // last face
                    glColor3f(1.0f, 0.0f, 0.0f);
                }
                else if (f[i].points[j].red != cred ||
                         f[i].points[j].green != cgreen ||
                         f[i].points[j].blue != cblue)
                {
                    cred = f[i].points[j].red;
                    cgreen = f[i].points[j].green;
                    cblue = f[i].points[j].blue;
                    glColor3f(cred, cgreen, cblue);
                }
                glVertex3dv(f[i].points[j].glcoords);
            }
        }
        glEnd();
    }

    glPopMatrix();

    //qglColor(Qt::red);
    //paintText();

    DGLWidget::paintGL();
}

void PointSetDisplay::paintText()
{
    QString s, n;

    s.append("rotate: ");
    n.setNum(_rotx);
    s.append(n);
    s.append(", ");
    n.setNum(_roty);
    s.append(n);
    s.append(", ");
    n.setNum(_rotz);
    s.append(n);
    renderText(10, 40, s);

    s.clear();
    s.append("pos: ");
    n.setNum(_posx);
    s.append(n);
    s.append(", ");
    n.setNum(_posy);
    s.append(n);
    s.append(", ");
    n.setNum(_posz);
    s.append(n);
    renderText(10, 20, s);
}

void PointSetDisplay::mousePressEvent(QMouseEvent *e)
{
    if (e->button() == Qt::LeftButton) {
        _mousex = e->globalX();
        _mousey = e->globalY();
        if (e->modifiers() & Qt::ShiftModifier)
            _zoom = true;
        else if (e->modifiers() & Qt::ControlModifier)
            _rotate = true;
        else
            _translate = true;
    }
    else if (e->button() == Qt::RightButton) {
        _mousex = e->globalX();
        _mousey = e->globalY();
        _rotate = true;
    }
}

void PointSetDisplay::mouseReleaseEvent(QMouseEvent *e)
{
    (void)e;
    _translate = _rotate = _zoom = false;
}

void PointSetDisplay::mouseMoveEvent(QMouseEvent *e)
{
    int deltax, deltay;

    // find how far the mouse moved (and invert the y axis to match the gl display's y axis)
    deltax = e->globalX() - _mousex;
    deltay = -(e->globalY() - _mousey);

    // remember the new position
    _mousex = e->globalX();
    _mousey = e->globalY();

    if (_zoom) {
        // move camera toward/away from pointset
        _posz += deltay / 50.0;
    }
    else if (_translate) {
        // translate camera
        _posx += (deltax / 50.0);
        _posy += (deltay / 50.0);
    }
    else if (_rotate) {
        // rotate camera around the pointset
        const float drotx = deltax / 1000.0f * 360.0f;
        const float droty = deltay / 1000.0f * 360.0f;

        _rotx += drotx;
        while (_rotx < -180.0f) _rotx += 360.0f;
        while (_rotx > 180.0f) _rotx -= 360.0f;
        _roty -= droty;
        if (_roty > 90.0f) _roty = 90.0f;
        if (_roty < -90.0f) _roty = -90.0f;
    }
}

void PointSetDisplay::initView()
{
    // find a z-position for the camera that will contain all the points
    // based on what the projection for the camera is, which is the
    // glfrustum() call made when the window resizes.  the formula
    // for the field of view angle from there is
    // fov*0.5 = arctan ((top-bottom)*0.5 / near)
    // fov = 2.0 * arctan((top - bottom) * 0.5 / near)

    const float fovy = 2.0 * atan((_bottom - _top) * 0.5 / _near);
    const float fovx = 2.0 * atan((_right - _left) * 0.5 / _near);

    // theta is the angle made from the center of the camera to one side
    float thetay = fovy / 2.0f;
    float thetax = fovx / 2.0f;

    // find the z position required to keep each point on screen,
    // and take the most extreme of those
    _posz = 0;
    for (int i = 0; i < _set->size(); ++i) {
        float zx, zy, y, x, z;
        const Point *p = &_set->points()[i];

        y = ABS(p->y()) + 1.0;
        // the distance from the point, when it is right at the edge of the camera
        zy = y / tan(thetay);

        x = ABS(p->x()) + 1.0;
        // the distance from the point, when it is right at the edge of the camera
        zx = x / tan(thetax);

        // use the larger distance, and make sure we are 1.0 away from everything
        // so they are past the front clipping plane
        z = MAX(1.0, MAX(zx, zy));

        // find the position where the camera would go relative to the point
        z += p->z();

        _posz = MIN(_posz, -z);
    }

    _posx = _posy = 0;

    _rotx = 0.0f;
    _roty = 0.0f;
    _rotz = 0.0f;

    _viewinit = true;
}

void PointSetDisplay::resetView()
{
    initView();
}

void PointSetDisplay::highlightLastFace(bool h)
{
    _hilightlast = h;
}

