#include "projectiondisplay.h"
#include "util.h"
#include <math.h>
#include <iostream>
#include <QFontMetrics>

#define PROJX(p, t) ((p).x())
#define PROJY(p, t) (MAX(MIN(((p).z() - t * (p).y()), Point::INF), -Point::INF))

const int ProjectionDisplay::DEFAULT_ANIM_SPEED_MS;
const float ProjectionDisplay::POINT_SIZE;
const double ProjectionDisplay::HILIGHT_POINT_TIME;
const double ProjectionDisplay::ZOOM_TIME;

ProjectionDisplay::ProjectionDisplay(QWidget *parent) :
        DGLWidget(parent, QGLFormat(QGL::DoubleBuffer)),
        _set(0),
        _anim_speed_ms(DEFAULT_ANIM_SPEED_MS),
        _state(STATE_IDLE),
        _auto_zoom(true)
{
    QFont font; // default font
    QFontMetrics qm = QFontMetrics(font);
    _text_height = qm.ascent();
}

ProjectionDisplay::~ProjectionDisplay()
{
    setPointSet(0);
}

void ProjectionDisplay::initializeGL()
{
    DGLWidget::initializeGL();

    qglClearColor(Qt::white);
    glPointSize(ProjectionDisplay::POINT_SIZE);
    glLineWidth(1.5f);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void ProjectionDisplay::resizeGL(int w, int h)
{
    DGLWidget::resizeGL(w, h);
    glViewport(0, 0, w, h);
}

void ProjectionDisplay::paintGrid()
{
    if (_zoom_bottom >= _zoom_top) return; // invalid display

    // draw lines based on the zoom level at fixed intervals/positions
    // can fade them in while zooming out, or fade them in while zooming in

    // so, take the area size.  divide by how many dark lines you want to show up in it.
    // then find the nearest power of two that fits between 2 of those lines.  multiples
    // of that number is the y-coords where the lines will be drawn.  halfway between can
    // fade the lines, depending on how close the power of 2 is to the actual desired distance
    // between the lines

    const double area = _zoom_top - _zoom_bottom;
    const double ideal_num_lines = 3.0;
    const double ideal_between_lines = area / ideal_num_lines;

    // find a power of two close to the ideal distance between lines
    double val = 1.0; // 2^x
    while (val+val < ideal_between_lines)
        val += val; // *2
    while (val >= ideal_between_lines)
        val /= 2.0;

    const double below = (_zoom_top < 0.0 ? (int)(_zoom_top / val) * val : 0.0);
    const double above = (_zoom_bottom > 0.0 ? (int)(_zoom_bottom / val) * val : 0.0);
    double y;

    // the alpha value for the lines between the main lines
    const double halfalpha = ABS(val+val - ideal_between_lines)/ideal_between_lines;

    glBegin(GL_LINES);

    // draw the main lines (they are full opacity)
    glColor4f(0.4f, 0.7f, 0.4f, 1.0);
    for (y = below; y > _zoom_bottom; y -= val) {
        glVertex2d(_zoom_left, y);
        glVertex2d(_zoom_right, y);
    }
    for (y = above; y < _zoom_top; y += val) {
        glVertex2d(_zoom_left, y);
        glVertex2d(_zoom_right, y);
    }

    // draw the lines halfway between them (they have varying opacity based on the zoom)
    glColor4f(0.4f, 0.7f, 0.4f, halfalpha);
    for (y = below - val/2.0; y > _zoom_bottom; y -= val) {
        glVertex2d(_zoom_left, y);
        glVertex2d(_zoom_right, y);
    }
    for (y = above + val/2.0; y < _zoom_top; y += val) {
        glVertex2d(_zoom_left, y);
        glVertex2d(_zoom_right, y);
    }

    glEnd();
}

void ProjectionDisplay::paintGL()
{
    glClear(GL_COLOR_BUFFER_BIT);

    if (_set) {
        Point *p;
        register int i, size;
        float cred, cgreen, cblue;

        p = _set->points();
        size = _set->size();

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluOrtho2D(_zoom_left, _zoom_right, _zoom_bottom, _zoom_top);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        paintGrid(); // paint grid lines to show what's going on while moving/zooming

        cred = cgreen = cblue = 0.0f;
        glColor3f(cred, cgreen, cblue);

        // draw the convex hull
        if (_bridgel && _bridger) {
            const Point *p;

            glColor3f(0.7f, 0.7f, 0.7f);

            // left hull
            glBegin(GL_LINE_STRIP);
            for (p = _bridgel; p != Point::NIL; p = p->prev)
                glVertex2d(PROJX(*p, _cur_t), PROJY(*p, _cur_t));
            glEnd();

            // right hull
            glBegin(GL_LINE_STRIP);
            for (p = _bridger; p != Point::NIL; p = p->next)
                glVertex2d(PROJX(*p, _cur_t), PROJY(*p, _cur_t));
            glEnd();

            // bridge
            glBegin(GL_LINES);
            glColor3f(0.2f, 0.2f, 0.8f);
            glVertex2d(PROJX(*_bridgel, _cur_t), PROJY(*_bridgel, _cur_t));
            glVertex2d(PROJX(*_bridger, _cur_t), PROJY(*_bridger, _cur_t));
            glEnd();

        }

        // draw the points
        glBegin(GL_POINTS);
        for (i = 0; i < size; ++i) {
            if (p+i == _hilite) {
                cred = 1.0f;
                cgreen = 0.0f;
                cblue = 0.0f;
                glColor3f(cred, cgreen, cblue);
            }
            else if (p[i].red != cred || p[i].green != cgreen || p[i].blue != cblue) {
                cred = p[i].red;
                cgreen = p[i].green;
                cblue = p[i].blue;
                glColor3f(cred, cgreen, cblue);
            }

            glVertex2d(PROJX(p[i], _cur_t), PROJY(p[i], _cur_t));
        }
        glEnd();

        // show the hilighted point
        if (_hilite && _state == STATE_HIGHLIGHTING) {
            glBegin(GL_LINE_LOOP);
            glColor3f(1.0f, 0.0f, 0.0f);

            const double px = PROJX(*_hilite, _cur_t);
            const double py = PROJY(*_hilite, _cur_t);
            const double remain = (_hilite_end - _hilite_cur) / HILIGHT_POINT_TIME;

            // draw a strinking box around the higlighted point
            glVertex2d(px - (px - _zoom_left) * remain, py - (py - _zoom_bottom) * remain);
            glVertex2d(px + (_zoom_right - px) * remain, py - (py - _zoom_bottom) * remain);
            glVertex2d(px + (_zoom_right - px) * remain, py + (_zoom_top - py) * remain);
            glVertex2d(px - (px - _zoom_left) * remain, py + (_zoom_top - py) * remain);
            glEnd();
        }
    }

    glColor3f(0.0f, 0.0f, 0.0f);
    QString t("t: ");
    if (_cur_t <= -Point::INF) t.append("-inf");
    else if (_cur_t >= Point::INF) t.append("inf");
    else t.append(QString::number(_cur_t));
    renderText(2, _text_height, t);

    DGLWidget::paintGL();
}

void ProjectionDisplay::setPointSet(PointSet *set)
{
    if (set) set->ref();

    if (_set) {
        _set->unref();
        _set = 0;
    }

    if (set) {
        _set = set;
        reset();
    }
}

void ProjectionDisplay::reset()
{
    assert(_set);

    _bridgel = _bridger = 0;

    _cur_t = _dest_t = -Point::INF;
    _hilite_cur = _hilite_end = 0.0;

    zoom();
}

void ProjectionDisplay::setAnimationSpeed(int ms)
{
    if (ms >= 0) _anim_speed_ms = ms;
}

void ProjectionDisplay::showHull(const Point *bridgel, const Point *bridger)
{
    assert(_set);
    _bridgel = bridgel;
    _bridger = bridger;
}

void ProjectionDisplay::highlightPoint(const Point *event)
{
    _hilite = event;
    _hilite_cur = 0.0;
    _hilite_end = HILIGHT_POINT_TIME;
}

void ProjectionDisplay::showTime(double t)
{
    assert(_set);

    _dest_t = t;

    _time_per_ms = (double)(_dest_t - _cur_t) / (double)_anim_speed_ms; // show the entire transition in this much time
}

void ProjectionDisplay::animate(int ms)
{
    DGLWidget::animate(ms);

    if (_hilite_cur < _hilite_end) {
        // hilite the point for which the event happened
        _hilite_cur = MIN(_hilite_end, _hilite_cur + ms);
        _state = STATE_HIGHLIGHTING;
    }
    else if (_cur_t != _dest_t) {
        // animate the points
        if (_cur_t < _dest_t)
            _cur_t = MIN(_cur_t + (_time_per_ms * ms), _dest_t);
        else
            _cur_t = MAX(_cur_t - (_time_per_ms * ms), _dest_t);

        if (_state != STATE_TRANSLATING) {
            //showOSDText("time travelling");
            _state = STATE_TRANSLATING;
        }

        if (_auto_zoom) zoom();
    }
    else if (_state != STATE_IDLE) {
        _state = STATE_IDLE;
        emit showingTime(_cur_t);
    }
}

void ProjectionDisplay::zoom()
{
    // find what the zoom should be
    double minx, maxx, miny, maxy;
    int i;
    Point *p;

    // find the size of the current time
    minx = miny = Point::INF;
    maxx = maxy = -Point::INF;
    for (i = 0, p = _set->points(); i < _set->size(); ++i) {
        minx = MIN(minx, PROJX(p[i], _cur_t));
        maxx = MAX(maxx, PROJX(p[i], _cur_t));
        miny = MIN(miny, PROJY(p[i], _cur_t));
        maxy = MAX(maxy, PROJY(p[i], _cur_t));
    }
    _zoom_left = minx - (maxx-minx)*0.025;
    _zoom_right = maxx + (maxx-minx)*0.025;
    _zoom_bottom = miny - (maxy-miny)*0.025;
    _zoom_top = maxy + (maxy-miny)*0.025;

    // don't let them get on top of each other
    if (_zoom_bottom > _zoom_top - 0.0001) {
        _zoom_bottom -= 0.0001/2.0;
        _zoom_top += 0.0001/2.0;
    }
}

void ProjectionDisplay::mousePressEvent(QMouseEvent *e)
{
    if (_zoom || _move) return;

    if (e->button() == Qt::LeftButton)
        _move = true;
    else if (e->button() == Qt::RightButton ||
             (e->button() == Qt::LeftButton && (e->modifiers() == Qt::ControlModifier)))
        _zoom = true;
    else
        return;

    _mousey = e->y();
}

void ProjectionDisplay::mouseReleaseEvent(QMouseEvent *e)
{
    (void)e;
    _zoom = _move = false;
}

void ProjectionDisplay::mouseMoveEvent(QMouseEvent *e)
{
    double d = (_zoom_top - _zoom_bottom) * 0.005 * (e->y() - _mousey);
    _zoom_top += (_zoom ? d : (_move ? d : 0));;
    _zoom_bottom -= (_zoom ? d : (_move ? -d : 0));
    _mousey = e->y();
}
